# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for processing spike trains (point processes) """
from typing import Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike


def psth(
    spikes: ArrayLike,
    binwidth: float,
    start: Optional[float] = None,
    stop: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute a PSTH of spike times.

    spikes: array or list of spike times
    binwidth: the bin duration (same units as spikes)
    start: the start of the observation interval. If None, the time of the first spike is used.
    stop: the end of the observation interval. If None, the time of the last spike plus one bin is used.

    Returns (spike counts, bin times)
    """
    spikes = np.asarray(spikes)
    t1 = start if start is not None else spikes.min()
    t2 = stop if stop is not None else (spikes.max() + binwidth)
    bins = np.arange(t1, t2, binwidth)
    counts, bins = np.histogram(spikes, bins)
    return counts, bins[:-1]


def rate(
    spikes: ArrayLike,
    binwidth: float,
    kernel: ArrayLike,
    start: Optional[float] = None,
    stop: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute a smoothed estimate of spike rate from spike times.

    spikes: array or list of spike times
    binwidth: the bin duration (same units as spikes)
    kernel: the smoothing kernel. Sampling interval needs to be same as binwidth.
    start: the start of the observation interval. If None, the time of the first spike is used.
    stop: the end of the observation interval. If None, the time of the last spike is used.

    Returns (rate estimate, bin times)
    """
    counts, bins = psth(spikes, binwidth, start, stop)
    return np.convolve(counts, kernel, mode="same"), bins



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Utility functions for scripts and modules """
import argparse
from functools import singledispatch

import numpy as np


def setup_log(log, debug=False):
    """Set up logging for a module.

    log: generate by calling e.g. `log = logging.getLogger("dlab.extracellular")`
    """
    import logging

    ch = logging.StreamHandler()
    formatter = logging.Formatter("%(message)s")
    loglevel = logging.DEBUG if debug else logging.INFO
    log.setLevel(loglevel)
    ch.setLevel(loglevel)
    ch.setFormatter(formatter)
    log.addHandler(ch)


class ParseKeyVal(argparse.Action):
    """argparse action for parsing -k key=value arguments

    Example: p.add_argument("-k", action=ParseKeyVal, default=dict(),
                            metavar="KEY=VALUE", dest="metadata")
    """

    def parse_value(self, value):
        import ast

        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value

    def __call__(self, parser, namespace, arg, option_string=None):
        kv = getattr(namespace, self.dest)
        if kv is None:
            kv = dict()
        if not arg.count("=") == 1:
            raise ValueError("%s argument badly formed; needs key=value" % arg)
        else:
            key, val = arg.split("=")
            kv[key] = self.parse_value(val)
        setattr(namespace, self.dest, kv)


@singledispatch
def json_serializable(val):
    """Serialize a value for the json module."""
    return str(val)


@json_serializable.register(np.generic)
def __js_numpy(val):
    """Used if *val* is an instance of a numpy scalar."""
    return val.item()


@json_serializable.register(np.ndarray)
def __js_numpy_arr(arr):
    """Used if *arr* is an instance of a numpy array."""
    return arr.tolist()


def all_same(seq):
    """If all elements of seq are equal, returns the value, otherwise None"""
    it = iter(seq)
    first = next(it)
    for e in it:
        if e != first:
            return None
    return first



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Signal processing functions """
from typing import Tuple

import numpy as np


def kernel(name: str, bandwidth: float, dt: float) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate a smoothing kernel function with bandwidth.

    name:      the name of the kernel. can be anything supported
               by scipy.signal.get_window, plus the following functions:
        epanech, biweight, triweight, cosinus, exponential

    bandwidth: the bandwidth of the kernel, in units of dt
    dt:        the time resolution of the kernel

    Returns:
    window:    the window function, normalized such that sum(w)*dt = 1.0
    grid:      the time support of the window, centered around 0.0

    From matlab code by Zhiyi Chi
    """
    if name in ("normal", "gaussian"):
        D = 3.75 * bandwidth
    elif name == "exponential":
        D = 4.75 * bandwidth
    else:
        D = bandwidth

    N = int(np.floor(D / dt))  # number of grid points in half the support
    G = (np.arange(1, 2 * N + 2) - 1 - N) * dt  # grid support
    xv = G / bandwidth

    if name in ("gaussian", "normal"):
        W = np.exp(-xv * xv / 2)
    elif name == "exponential":
        xv = np.minimum(xv, 0)
        W = np.absolute(xv) * np.exp(xv)
    elif name in ("biweight", "quartic"):
        W = np.maximum(0, 1 - xv * xv) ** 2
    elif name == "triweight":
        W = np.maximum(0, 1 - xv * xv) ** 3
    elif name == "cosinus":
        W = np.cos(np.pi * xv / 2) * (np.absolute(xv) <= 1)
    elif name == "epanech":
        W = np.maximum(0, 1 - xv * xv)
    elif name == "hanning":
        W = np.hanning(1 + 2 * N)
    else:
        from scipy.signal import get_window

        W = get_window(name, 1 + 2 * N)

    return W / (W.sum() * dt), G



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Utilities for extracellular experiments """
import json
import logging
import re

logging.getLogger(__name__).addHandler(logging.NullHandler())


def entry_time(entry):
    """Return the timestamp of an entry as a floating point number"""
    from arf import timestamp_to_float

    return timestamp_to_float(entry.attrs["timestamp"])


def entry_datetime(entry):
    """Return the timestamp of an entry as a floating point number"""
    from arf import timestamp_to_datetime

    return timestamp_to_datetime(entry.attrs["timestamp"])


def iter_entries(data_file):
    """Iterate through the entries in an arf file in order of time"""
    return enumerate(sorted(data_file.values(), key=entry_time))


def find_stim_dset(entry):
    """Returns the first dataset that matches 'Network_Events.*_TEXT'"""
    rex = re.compile(r"Network_Events-.*?TEXT")
    for name in entry:
        if rex.match(name) is not None:
            logging.debug("  - stim log dataset: %s", name)
            return entry[name]


def entry_metadata(entry):
    """Extracts metadata from an entry in an oeaudio-present experiment ARF file.

    Metadata are passed to open-ephys through the network events socket as a
    json-encoded dictionary. There should be one and only one metadata message
    per entry, so only the first is returned.

    """
    re_metadata = re.compile(r"metadata: (\{.*\})")
    stims = find_stim_dset(entry)
    for row in stims:
        message = row["message"].decode("utf-8")
        m = re_metadata.match(message)
        try:
            metadata = json.loads(m.group(1))
        except (AttributeError, json.JSONDecodeError):
            pass
        else:
            metadata.update(name=entry.name, sampling_rate=stims.attrs["sampling_rate"])
            return metadata



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Functions for interfacing with the neurobank repository """
import logging
from typing import Dict, Union
from urllib.parse import urlparse
import json

from anyio import Path, create_task_group
from httpx import AsyncClient, HTTPStatusError

from dlab import cache
from nbank import registry, util

logging.getLogger(__name__).addHandler(logging.NullHandler())
default_registry = registry.default_registry()


async def find_resource(
    session: AsyncClient,
    registry_url: str,
    resource_id: str,
    alt_base: Union[Path, str, None] = None,
    no_download: bool = False,
) -> Path:
    """Locate a neurobank resource

    This function will try to locate resources from the following locations:
    - a local directory (alt_base, if set),
    - a local neurobank archive (using alt_base, if provided),
    - a local cache
    - a remote HTTP archive (caching the file for local access later)

    It will raise a ValueError if the resource does not exist and a
    FileNotFoundError if the resource cannot be located.

    """
    url, params = registry.get_locations(registry_url, resource_id)
    if alt_base is not None:
        stem = Path(alt_base) / resource_id
        path = await resolve_local_path(stem)
        if path is not None:
            logging.debug("%s: found in alt_base", resource_id)
            return path
    response = await session.get(url, params=params)
    response.raise_for_status()
    locations = response.json()
    # search for local files
    for loc in locations:
        if loc["scheme"] not in registry._local_schemes:
            continue
        path = await resolve_local_path(Path(util.parse_location(loc, alt_base)))
        if path is not None:
            logging.debug("%s: found in local repository", resource_id)
            return path
    # search remote locations
    for loc in locations:
        if loc["scheme"] in registry._local_schemes:
            continue
        url = util.parse_location(loc)
        try:
            return await fetch_resource(session, url, resource_id, no_download)
        except FileNotFoundError:
            pass
    # all locations failed; raise an error
    raise FileNotFoundError(f"{resource_id}: unable to locate file")


async def resolve_local_path(stem: Path) -> Path:
    """Find a local resource based on a path or path stem.

    Because resource names often don't have extensions, a local filesystem
    search may be needed to resolve the full path.

    """
    if await stem.exists():
        return stem
    async for path in stem.parent.glob(f"{stem.name}.*"):
        return path


async def fetch_resource(
    session: AsyncClient,
    url: str,
    resource_id: str,  # this could be parsed out of the url
    no_download: bool = False,
) -> Path:
    """Fetch a downloadable resource from the registry.

    The file will be cached locally. Returns the path of the file.
    """
    target = await cache.locate(resource_id, urlparse(url).netloc)
    if await target.exists():
        logging.debug("%s: found in local cache", resource_id)
        return target
    elif no_download:
        raise FileNotFoundError(f"{url}: resource not found in local cache")
    logging.debug("%s: fetching from registry", resource_id)
    async with session.stream("GET", url) as response:
        if response.status_code == 404:
            raise FileNotFoundError(f"{url}: resource not found")
        elif response.status_code == 415:
            # should be possible to get the error message from registry?
            raise FileNotFoundError(f"{url}: this data type is not downloadable")
        async with await target.open("wb") as fp:
            async for data in response.aiter_bytes():
                await fp.write(data)
    return target


async def fetch_metadata(
    session: AsyncClient, registry_url: str, resource_id: str
) -> Dict:
    """Fetch metadata for a resource. Results will be cached locally."""
    target = await cache.locate(resource_id, urlparse(registry_url).netloc)
    if await target.exists():
        logging.debug("%s: metadata found in local cache", resource_id)
        return json.loads(await target.read_text())
    else:
        logging.debug("%s: fetching metadata from registry", resource_id)
        url, params = registry.get_resource(registry_url, resource_id)
        response = await session.get(url, params=params)
        response.raise_for_status()
        data = response.json()
        await target.write_text(json.dumps(data))
        return data


def add_registry_argument(parser, dest="registry_url"):
    """Add a registry argument to an argparse parser"""
    parser.add_argument(
        "-r",
        dest=dest,
        help="URL of the registry service. "
        "Default is to use the environment variable '%s'" % registry._env_registry,
        default=default_registry,
    )


async def main(argv=None):
    import argparse

    p = argparse.ArgumentParser(description="locate neurobank resources ")
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    add_registry_argument(p)
    p.add_argument(
        "-b",
        "--base",
        type=Path,
        help="set an alternative base directory to search for local resources",
    )
    p.add_argument(
        "--clear-cache",
        action="store_true",
        help="clear the contents of the local cache",
    )
    p.add_argument("id", help="identifier(s) of the resource(s) to locate", nargs="*")
    args = p.parse_args(argv)
    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if args.debug else logging.INFO
    )

    if args.clear_cache:
        await cache.clear(urlparse(args.registry_url).netloc)

    async def _find_and_show(session, id):
        try:
            path = await find_resource(
                session, args.registry_url, id, alt_base=args.base
            )
        except HTTPStatusError:
            logging.info("%s: no such resource", id)
        except FileNotFoundError:
            logging.info("%s: not found, unable to download", id)
        else:
            logging.info("%s: %s", id, path)

    headers = {"Accept": "application/json"}
    async with AsyncClient(
        timeout=None, headers=headers
    ) as session, create_task_group() as tg:
        for id in args.id:
            tg.start_soon(_find_and_show, session, id)


if __name__ == "__main__":
    from anyio import run

    run(main)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Local cache, primarily for storing neurobank resources """
import logging

from anyio import Path, run_process
import appdirs

APP_NAME = "dlab"
APP_AUTHOR = "melizalab"
user_dir = appdirs.user_cache_dir(APP_NAME, APP_AUTHOR)
logging.getLogger(__name__).addHandler(logging.NullHandler())


async def locate(name: str, subdir: str) -> Path:
    """Return any anyio.Path for a cached resource, creating subdir in the cache if needed"""
    cache_dir = Path(user_dir) / subdir
    await cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir / name


async def clear(subdir: str) -> None:
    """Clear the contents of the cache"""
    cache_dir = Path(user_dir) / subdir
    cmd = ["rm", "-rf", str(await cache_dir.resolve())]
    logging.debug("clearing local cache dir %s: '%s'", cache_dir, " ".join(cmd))
    _ = await run_process(cmd)
